#
# @(#)Whois.pm
#
# DESCRIPTION
#
# Author:	Leo Kim. <leo@florida.sarang.net>
# Date:		04/16/2001
# Revision:	$Revision: 1.1.1.1 $
# Last Update:	$Date: 2002/09/03 09:15:12 $
#

#
#  Copyright (C) 2001 Leo Kim.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

package Whois;

use strict;
use Socket qw(:DEFAULT :crlf);
use Date::Parse;
use IO::Handle;
use Whois;

##############################################################################
# Module preparation
#
BEGIN {
    use Exporter ();
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    # set the version for version checking
    $VERSION = do { my @r = (q$Revision: 1.1.1.1 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

    # ISA
    @ISA = qw(Exporter);

    # Symbols to export by default
    @EXPORT = qw();

    # Symbos to export on request
    @EXPORT_OK = qw();

    # define names for set of symbols. (eg: TAG => [ qw(name1 name2) ])
    %EXPORT_TAGS = ();
}

##############################################################################
# Constants
#
use constant QNICHOST_TAIL	=> ".whois-servers.net";
use constant REX_REFERRAL       => qr/^\s+Whois\s+Server:\s([\w\-\.]+)/i;

##############################################################################
# Class variables
#
my $verbose		= 0;
my $debug		= 0;
my $max_conn_try	= 3;
my $rex_tld		= qr/\.([^\.]+)$/;
my $nuniq_rex		= qr/^To\s+single\s+out\s+one\s+record,/;
my $np			= [ qr/^\s*No\s+match\s+for\s+/,
			    qr/^\s+\s*̸\s+ϵǾ\s+\s+ʽϴ/, ];

##############################################################################
# Debug print
#
# DEBUG::_dprintf $level, $fmt, ...
#
# $level = 0 : no debug output
#          1 : program execution flow
#          2 : network connection
#          3 : whois parser
#          4 : buffer contents
#
sub DEBUG::_dprintf($@) {
    my $level = shift;

    printf STDERR @_ if ($level <= $debug);
}

##############################################################################
# Verbose print
#
# DEBUG::_vprintf $fmt, ...
#
sub DEBUG::_vprintf(@) {
    return unless $verbose;
    printf STDOUT @_;
}

##############################################################################
# Constructor
#
sub new($) {
    my $proto = shift;
    my ($domain, $opt) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};

    # options
    $self->{VERBOSE}		= $verbose;
    $self->{DEBUG}		= $debug;
    $self->{MAX_CONN_TRY}	= $max_conn_try;

    if (defined $opt) {
	foreach my $k (keys %{$opt}) {
	    $self->{uc($k)} = $opt->{$k};
	}
    }

    $verbose      = $self->{VERBOSE};
    $debug        = $self->{DEBUG};
    $max_conn_try = $self->{MAX_CONN_TRY};

    # private data
    $self->{OK}			= undef;
    $self->{DOMAIN}		= $domain;
    $self->{DOMAIN_HANDLE}	= undef;
    $self->{REGISTRANT}		= undef;
    $self->{ADDRESS}		= undef;
    $self->{COUNTRY}		= undef;
    $self->{ADMIN}		= undef;
    $self->{ADMIN_HANDLE}	= undef;
    $self->{EMAIL}		= undef;
    $self->{PHONE}		= undef;
    $self->{FAX}		= undef;
    $self->{CREATED}		= undef;
    $self->{UPDATED}		= undef;
    $self->{EXPIRES}		= undef;
    $self->{WHOIS_SERVER}	= undef;
    $self->{REFERRAL}		= undef;
    $self->{NAME_SERVERS}	= [];

    # read Whois database
    my $not_found = 0;
    my @lines;

    if ($domain =~ /$rex_tld/) {
	my $tld = $1;
	my $server = $tld . QNICHOST_TAIL;

	# connect to the whois server
	return undef unless _whois($domain, $server, 1, \@lines);

	# if the result is not unique, try again with '=' prefix.
	my $unique = 1;

	foreach (@lines) {
	    if (/$nuniq_rex/) {
		$unique = undef;
		last;
	    }
	}
	unless ($unique) {
	    @lines = ();
	    return undef unless _whois('='.$domain, $server, 1, \@lines);
	}

	# debug
	DEBUG::_dprintf 4, "# :::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n";
	do { chomp; DEBUG::_dprintf 4, "! %s\n", $_; } foreach (@lines);
	    
	# search negative response
      END_FOR:
	foreach (@$np) {
	    my $rex = $_;
	    
	    foreach (@lines) {
		if (/$rex/) {
		    DEBUG::_vprintf "%s: No such domain\n", $domain;
		    $not_found = 1;
		    last END_FOR;
		}
	    }
	}
    }
    else {
	return undef;
    }

    # ok
    $self->{OK} = ($not_found ? 0 : 1);

    # parse
    unless ($not_found) {
	my $rv;

	$rv = _parse_whois(\@lines);
	if ($rv) {
	    foreach my $k (keys %{$rv}) {
		$self->{uc($k)} = $rv->{$k};
	    }
	}
	else {
	    $self->{OK} = 0;
	}
    }

    # done
    bless($self, $class);

    return $self;
}

##############################################################################
# Get Whois information from the server
#
# whois $domain, $server, $flag, $lines;
# 
# $domain : domain name
# $server : whois server
# $flag   : 1 if recursive search
# $lines  : line array reference
#
sub _whois($$$$) {
    my ($domain, $server, $flag, $lines) = @_;

    DEBUG::_dprintf 1, "# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n";
    DEBUG::_dprintf 1, "# WHOIS Search (Recursive = %s)\n", ($flag ? "True" : "False");
    DEBUG::_dprintf 1, "# domain = %s\n", $domain;
    DEBUG::_dprintf 1, "# server = %s\n", $server;

    # resolve the server name to IP address
    my $proto = getprotobyname('tcp');
    my $port  = getservbyname('whois', 'tcp');
    my $iaddr = gethostbyname($server);

    unless ($iaddr) {
	DEBUG::_vprintf "? %s: No such host\n", $server;
	return undef;
    }

    my $sin = sockaddr_in($port, $iaddr);
    my $sock;
    my @lines;

    # try to connect, and read contents
    my $try = $max_conn_try;
    while ($try > 0) {
	close $sock if $sock;
	socket(SOCK, PF_INET, SOCK_STREAM, $proto) 
	    or die "socket: $!\n";
	$sock = \*SOCK;

	# connect
	unless (connect($sock, $sin)) {
	    $try--;
	    DEBUG::_dprintf 2, "# %s: fail to connect (%d)\n", $server, ($max_conn_try - $try);
	    sleep(($max_conn_try - $try) * 2);
	    next;
	}

	chomp(my $now = localtime);
	DEBUG::_dprintf 2, "# %s - %s: connected\n", $now, $server;

	# prepare buffered input/output streams
	my ($fin, $fout);

	$fin  = new IO::Handle;
	$fout = new IO::Handle;

	die "fin->fdopen: $!" unless $fin->fdopen(fileno(SOCK), "r");
	die "fout->fdopen: $!" unless $fout->fdopen(fileno(SOCK), "w");

	# send whois query
	$fout->print($domain . CRLF);
	$fout->flush;

	# read response
	@lines = $fin->getlines;

	# close streams
	$fin->close;
	$fout->close;

	# check empty contents
	if ($#lines < 0) {
	    $try--;
	    DEBUG::_dprintf 2, "# %s: empty contents (%d)\n", $server, ($max_conn_try - $try);
	    sleep(($max_conn_try - $try) * 2);
	    next;
	}

	# done
	last;
    }
    close $sock if $sock;

    # return no result if the max tries has been reached.
    return undef if $try < 1;

    # check if the response is a referral
    my $referral = undef;
    my $rex      = REX_REFERRAL;

    if ($flag) {
	my $got_domain_name = undef;

	foreach (@lines) {
	    chomp;

	    if (/^\s+Domain\s+Name:/) {
		$got_domain_name = 1;
		next;
	    }
	    
	    if ($got_domain_name && /$rex/) {
		$referral = $1;
		last;
	    }
	}
    }
    
    # add lines
    foreach (@lines) {
	$lines->[$#$lines + 1] = $_;
    }

    if ($referral) { # jump to the referral
	DEBUG::_dprintf 2, "# %s: jump to the referral\n", $referral;
	
	# second lookup (non-recursive)
	_whois($domain, $referral, 0, $lines);
    }

    return 1;
}

##############################################################################
# Access Methods
#
sub ok {
    my $self = shift;
    return $self->{OK};
}

sub domain {
    my $self = shift;
    return $self->{DOMAIN};
}

sub domain_handle {
    my $self = shift;
    return $self->{DOMAIN_HANDLE};
}

sub registrant {
    my $self = shift;
    return $self->{REGISTRANT};
}

sub address {
    my $self = shift;
    return $self->{ADDRESS};
}

sub country {
    my $self = shift;
    return $self->{COUNTRY};
}

sub admin {
    my $self = shift;
    return $self->{ADMIN};
}

sub admin_handle {
    my $self = shift;
    return $self->{ADMIN_HANDLE};
}

sub email {
    my $self = shift;
    return $self->{EMAIL};
}

sub phone {
    my $self = shift;
    return $self->{PHONE};
}

sub fax {
    my $self = shift;
    return $self->{FAX};
}

sub created {
    my $self = shift;
    return $self->{CREATED};
}

sub updated {
    my $self = shift;
    return $self->{UPDATED};
}

sub expires {
    my $self = shift;
    return $self->{EXPIRES};
}

sub registrar {
    my $self = shift;
    return $self->{REGISTRAR};
}

sub whois_server {
    my $self = shift;
    return $self->{WHOIS_SERVER};
}

sub referral {
    my $self = shift;
    return $self->{REFERRAL};
}

sub name_servers {
    my $self = shift;
    return $self->{NAME_SERVERS};
}

##############################################################################
# Trim a line
#
sub _trim($) {
    $_ = shift;

    s/^\s+//;
    s/\s+$//;

    return $_;
}

##############################################################################
# Frequently Used Regular Expressions for Parsers
#
my $rex_domain		= qr/^\s+Domain\s+Name:\s+([\w\-\.]+)/;
my $rex_registrar	= qr/^\s+Registrar:\s+(.+)/;
my $rex_server		= qr/^\s+Whois\s+Server:\s+([\w\-\.]+)/;
my $rex_referral	= qr/^\s+Referral\s+URL:\s+(.+)/;
my $rex_registrant	= qr/^Registrant:/;
my $rex_admin_contact	= qr/^\s+Administrative(?:(?:\s+Contact)|,)?/;
my $rex_email		= qr/(\S+\@\S+)/;
my $rex_nsi_1		= qr/^([^\(]+)\((.+)\)/;
my $rex_nsi_dm		= qr/^\s+Domain\s+Name\:\s+/;
my $rex_nsi_country	= qr/^\s+(..)\s*$/;
my $rex_nsi_admin	= qr/^\s+([^\(]+)\((.+)\)\s+(\S+\@\S+)/;
my $rex_nsi_fax		= qr/^\s+FAX[^\s]+\s*([\d\-\.\s]*[\d\-\.]+)/i;
my $rex_nsi_fax_phone	= qr/^\s+([\d\-\.\s]*[\d\-\.]+)(?:\s+\(FAX\)\s+([\d\-\.\s]*[0-9\-\.]+))?/;
my $rex_nsi_created	= qr/^\s+Record\s+created\s+on\s+(.+)\./i;
my $rex_nsi_updated	= qr/^\s+Record\s+last\s+updated\s+on\s+(.+)\./;
my $rex_nsi_expires	= qr/^\s+Record\s+expires\s+on\s+(.+)\./;
my $rex_nsi_ns_begin	= qr/^\s+Domain\s+servers\s+in\s+listed\s+order:/;
my $rex_nsi_ns		= qr/^\s+([\w\-\.]+)(?:\s+([\d\.A-Fa-f\:]+))?/;
my $rex_krnic_domain	= qr/^\s+̸\s+:\s+([\w\-\.]+)/;
my $rex_krnic_org	= qr/^.ι\s+:\s+(.+)/;
my $rex_krnic_address	= qr/^ּ\s+:\s+(.+)/;
my $rex_krnic_zip	= qr/^ȣ\s+:\s+(.+)/;
my $rex_krnic_admin	= qr/^\s+:\s+(.+)/;
my $rex_krnic_phone	= qr/^ȭȣ\s+:\s+(.+)/;
my $rex_krnic_email	= qr/^ڿּ\s+:\s+(\S+\@\S+)/;
my $rex_krnic_created	= qr/^\s+:\s+(\d{4})(\d{2})(\d{2})/;
my $rex_krnic_updated	= qr/^\s+\s+:\s+(\d{4})(\d{2})(\d{2})/;
my $rex_krnic_expires	= qr/^\s+\s+:\s+(\d{4})(\d{2})(\d{2})/;
my $rex_krnic_ns_begin	= qr/^1\s+Ӽ/;
my $rex_krnic_ns_name	= qr/^\s+ȣƮ\s+̸\s+:\s+([\w\-\.]+)/;
my $rex_krnic_ns_ip	= qr/^\s+IP\s+ּ\s+:(?:\s*([\d\.A-Fa-f\:]+))?/;
my $rex_netpia_admin	= qr/^\s+(.+)\s+(\d+\-\d+\-\d+)/;
my $rex_netpia_created	= qr/^\s+Record\s+created\s+on\.+\:\s+([^\.]+)\./i;
my $rex_netpia_updated	= qr/^\s+Record\s+last\s+updated\s+on\.+\:\s+([^\.]+)\./i;
my $rex_netpia_expires	= qr/^\s+Record\s+expires\s+on\.+\:\s+([^\.]+)\./i;
my $rex_netpia_ns_begin	= qr/^\s+Domain\s+Name\s+Servers\s+in\s+listed\s+order/i;
my $rex_netpia_done	= qr/^Please\s+register/;
my $rex_han_admin	= qr/(.+)\s+(\S+\@\S+)/;
my $rex_tierra_phone	= qr/^\s+([\d\-]+)/;
my $rex_tierra_fax	= qr/^\s+([\d\-]+)\s+\[fax\]/;
my $rex_tierra_created	= qr/^\s+Domain\s+created\s+on\s+(.+)/;
my $rex_tierra_updated	= qr/^\s+Last\s+updated\s+on\s+(.+)/;
my $rex_tierra_expires	= qr/^\s+Domain\s+expires\s+on\s+(.+)/;
my $rex_yesnic_reg	= qr/^\:\:Registrant\:\:/;
my $rex_yesnic_org	= qr/^\s+Name\s+\:\s+(.+)/;
my $rex_yesnic_address	= qr/^\s+Address\s+\:\s+(.+)/;
my $rex_yesnic_zip	= qr/^\s+Zipcode\s+\:\s+([\d\-]+)/;
my $rex_yesnic_country	= qr/^\s+Nation\s+\:\s+(..)/;
my $rex_yesnic_adm	= qr/^\:\:Administrative\s+Contact\:\:/;
my $rex_yesnic_admin	= qr/^\s+Name\s+\:\s+(.+)/;
my $rex_yesnic_email	= qr/^\s+Email\s+\:\s+(\S+\@\S+)/;
my $rex_yesnic_phone	= qr/^\s+Tel\s+\:\s+([\d\-]+)/;
my $rex_yesnic_fax	= qr/^\s+Fax\s+\:\s+([\d\-]+)/;
my $rex_yesnic_ns_begin	= qr/^\:\:Name\s+Servers\:\:/;
my $rex_yesnic_date	= qr/^\:\:Dates/;
my $rex_yesnic_created	= qr/^\s+Created\s+Date\s+(.+)/;
my $rex_yesnic_updated	= qr/^\s+Updated\s+Date\s+(.+)/;
my $rex_yesnic_expires	= qr/^\s+Valid\s+Date\s+(.+)/;

##############################################################################
# Whois parsers
#
# The result of Whois database varies from servers to servers.
# So we need Whois parsers as many as necessary.
#
# To register a new Whois parser:
# 
# 1. Create a new entry in $wp above with a proper parser subroutine prototype.
# 2. $wp fileds are
#      clue =>      : clue to distinguish the parser
#      parser =>    : parser subroutine
#      registrar => : for comment
# 3. The parser subroutine returns a reference or `undef' if errors.
# 4. The reference should contain following fields.
#      domain =>        : domain name
#      domain_handle => : domain name handle used in Network Solutions
#      registrant =>    : registrant name
#      address =>       : registrant address
#      country =>       : registrant country
#      admin =>         : administrator name
#      admin_handle =>  : administrator handle used in Network Solutions
#      email =>         : administrator email address
#      phone =>         : administrator phone number
#      fax =>           : administrator fax number
#      created =>       : domain created date (Unix time)
#      updated =>       : domain last updated date (Unix time)
#      expires =>       : domain expiration date (Unix time)
#      registrar =>     : registrar
#      whois_server =>  : whois server hostname
#      referral =>      : whois server referral URL (home page URL, ...)
#      name_servers =>  : array reference of name server hash references
#
#      *** The fields are all optional.
# 5. Use already compiled regular expressions if possible.
#
my $wp;

sub _p_netpia;
sub _p_hangang;
sub _p_nsi;
sub _p_tierranet;
sub _p_tucows;
sub _p_yesnic;
sub _p_krnic;

push @$wp, { clue => qr/^Welcome\s+to\s+NETPIA\.COM.s\s+WHOIS\s+Server/,
	     parser => \&_p_netpia,
	     registrar => 'NETPIA.COM, INC.'};

push @$wp, { clue => qr/^Welcome\s+to\s+DOREGI\.COM.s\s+WHOIS\s+data\s+service/,
	     parser => \&_p_hangang,
	     registrar => 'HANGANG SYSTEMS, INC.' };

push @$wp, { clue => qr/^The\s+Data\s+in\s+Network\s+Solutions.\s+WHOIS\s+data/,
	     parser => \&_p_nsi,
	     registrar => 'NETWORK SOLUTIONS, INC.' };

push @$wp, { clue => qr/^Register\s+your\s+domain\s+today\s+at\s+http:\/\/www\.domaindiscover\.com/,
	     parser => \&_p_tierranet,
	     registrar => 'TIERRANET, INC.' };

push @$wp, { clue => qr/^\s+Registrar:\s+TUCOWS/,
	     parser => \&_p_tucows,
	     registrar => 'TUCOWS.COM, INC.' };

push @$wp, { clue => qr/^\s+Registrar:\s+YESNIC\s+CO\./,
	     parser => \&_p_yesnic,
	     registrar => 'YESNIC CO. LTD.' };

push @$wp, { clue => qr/^Korea\s+Internet\s+Information\s+Service/,
	     parser => \&_p_krnic,
	     registrar => 'KRNIC' };


##############################################################################
# Parse Whois output
#
sub _parse_whois($) {
    my $lines = shift;
    
    # find a parser using clues.
    my $p_idx = undef;
    for (my $i = 0; $i <= $#$wp && !defined($p_idx); $i++) {
	my $rex = $wp->[$i]->{clue};

	foreach (@$lines) {
	    if (/$rex/) {
		$p_idx = $i;
		last;
	    }
	}
    }

    unless (defined($p_idx)) {
	DEBUG::_dprintf 3, "%% no parser available\n";
	return undef;
    }

    DEBUG::_dprintf 3, "%% parser found for '%s'\n", $wp->[$p_idx]->{registrar};

    my $parser = $wp->[$p_idx]->{parser};

    return &{$parser}($lines);
}


##############################################################################
# PARSER: NSI
#
sub _p_nsi {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_registrant/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # first line after Registrant:
	      if (/$rex_nsi_1/) {
		  $registrant = _trim($1);
		  $domain_handle = $2;
		  $state = 2;
	      }
	      last SWITCH;
	  }

	  if ($state == 2) { # address line 1
	      $address = _trim($_);
	      $state = 3;
	      last SWITCH;
	  }

	  if ($state == 3) { # address line 2
	      $address .= "\$" . _trim($_);
	      $state = 4;
	      last SWITCH;
	  }

	  if ($state == 4) { # address line 3, country code or none
	      if (/$rex_nsi_dm/) {
		  $state = 6;
		  last SWITCH;
	      }

	      if (/$rex_nsi_country/) {
		  $country = $1;
		  $state = 5;
	      }
	      elsif (/^.+\S+.+$/) {
		  $address .= "\$" . _trim($_);
		  $state = 5;
	      }
	      last SWITCH;
	  }

	  if ($state == 5) { # country code or more address lines
	      unless ($country) {
		  if (/$rex_nsi_country/) {
		      $country = $1;
		      $state = 6;
		  }
		  else {
		      $address .= "\$" . _trim($_);
		  }
	      }
	      last SWITCH;
	  }

	  if ($state == 6) { # Administrative contact
	      $state = 7 if (/$rex_admin_contact/);
	      last SWITCH;
	  }
	  
	  if ($state == 7) { # admin information
	      if (/$rex_nsi_admin/) {
		  $admin = _trim($1);
		  $admin_handle = $2;
		  $email = $3;
	      }
	      elsif (/$rex_nsi_fax/) {
		  $fax = $1;
	      }
	      elsif (/$rex_nsi_fax_phone/) {
		  $phone = $1;
		  $fax = $2 if $2;
	      }
	      elsif (/$rex_nsi_updated/) {
		  $updated = str2time($1);
	      }
	      elsif (/$rex_nsi_expires/) {
		  $expires = str2time($1);
	      }
	      elsif (/$rex_nsi_created/) {
		  $created = str2time($1);
		  $state = 8;
	      }
	      last SWITCH;
	  }

	  if ($state == 8) { # begin name server list
	      if (/$rex_nsi_ns_begin/) {
		  $state = 9;
	      }
	      last SWITCH;
	  }

	  if ($state == 9) { # name servers
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;

		  $nmservers[$#nmservers+1] = $ref;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: KRNIC
#
sub _p_krnic {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;
    my $fmt = '%04d-%02d-%02d';

    foreach (@$lines) {
      SWITCH: {
	  if ($state == 0) {
	    SW:
	      {
		  $state      = 1,       last SW if /$rex_krnic_ns_begin/;
		  $domain     = $1,      last SW if /$rex_krnic_domain/;
		  $registrant = $1,      last SW if /$rex_krnic_org/;
		  $address    = $1,      last SW if /$rex_krnic_address/;
		  $address   .= "\$".$1, last SW if /$rex_krnic_zip/;
		  $admin      = $1,      last SW if /$rex_krnic_admin/;
		  $phone      = $1,      last SW if /$rex_krnic_phone/;
		  $email      = $1,      last SW if /$rex_krnic_email/;
		  $created    = str2time(sprintf($fmt, $1, $2, $3)), last SW if /$rex_krnic_created/;
		  $updated    = str2time(sprintf($fmt, $1, $2, $3)), last SW if /$rex_krnic_updated/;
		  $expires    = str2time(sprintf($fmt, $1, $2, $3)), last SW if /$rex_krnic_expires/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # name server hostname
	      if (/$rex_krnic_ns_name/) {
		  $nmservers[$#nmservers+1] = { host => $1 };
		  $state = 2;
	      }
	      last SWITCH;
	  }

	  if ($state == 2) { # name server ip address
	      if (/$rex_krnic_ns_ip/) {
		  $nmservers[$#nmservers] = { ip => $1 } if $1;
		  $state = 1;
	      }
	      last SWITCH;
	  }
      }
    }

    # hard-wired values
    $country = "KR";
    $registrar = "KRNIC";
    $whois_server = "whois.krnic.net";
    $referral = "www.krnic.net";

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: NETPIA
#
sub _p_netpia {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_registrant/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # first line after Registrant:
	      $registrant = _trim($_);
	      $state = 2;
	      last SWITCH;
	  }

	  if ($state == 2) { # address line
	      $address = _trim($_);
	      $state = 3;
	      last SWITCH;
	  }

	  if ($state == 3) { # Administrative contact
	      $state = 4 if /$rex_admin_contact/;
	      last SWITCH;
	  }
	  
	  if ($state == 4) { # admin name and phone number
	      if (/$rex_netpia_admin/) {
		  $admin = $1;
		  $phone = $2;
		  $state = 5;
	      }
	      last SWITCH;
	  }

	  if ($state == 5) { # email
	      if (/$rex_email/) {
		  $email = $1;
		  $state = 6;
	      }
	      last SWITCH;
	  }

	  if ($state == 6) { # dates
	    SWITCH_1: 
	      {
		  $created = str2time($1), last SWITCH_1 if /$rex_netpia_created/;
		  $updated = str2time($1), last SWITCH_1 if /$rex_netpia_updated/;
		  $expires = str2time($1), last SWITCH_1 if /$rex_netpia_expires/;
		  $state   = 7,            last SWITCH_1 if /$rex_netpia_ns_begin/;
	      }
	      last SWITCH;
	  }

	  if ($state == 7) { # name servers
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;

		  $nmservers[$#nmservers+1] = $ref;
	      }
	      elsif (/$rex_netpia_done/) {
		  $state = 8;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: HANGANG
#
sub _p_hangang {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_registrant/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # first line after Registrant:
	      $registrant = _trim($_);
	      $state = 2;
	      last SWITCH;
	  }

	  if ($state == 2) { # address
	      $address = _trim($_);
	      $state = 3;
	      last SWITCH;
	  }

	  if ($state == 3) { # Administrative,
	      $state = 4 if /$rex_admin_contact/;
	      last SWITCH;
	  }
	  
	  if ($state == 4) { # admin name and email
	      if (/$rex_han_admin/) {
		  $admin = _trim($1);
		  $email = $2;
		  $state = 5;
	      }
	      last SWITCH;
	  }

	  if ($state == 5) { # telephone (and fax)
	      if (/$rex_nsi_fax_phone/) {
		  $phone = $1;
		  $fax = $2 if $2;
		  $state = 6;
	      }
	      last SWITCH;
	  }

	  if ($state == 6) { # dates
	    SWITCH_1:
	      {
		  $created = str2time($1), last SWITCH_1 if /$rex_netpia_created/;
		  $updated = str2time($1), last SWITCH_1 if /$rex_netpia_updated/;
		  $expires = str2time($1), last SWITCH_1 if /$rex_netpia_expires/;
		  $state   = 7,            last SWITCH_1 if /$rex_netpia_ns_begin/;
	      }
	      
	      last SWITCH;
	  }

	  if ($state == 7) { # name servers
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;
		  
		  $nmservers[$#nmservers+1] = $ref;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: TIERRANET
#
sub _p_tierranet {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_registrant/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # first line after Registrant:
	      $registrant = _trim($_);
	      $state = 2;
	      last SWITCH;
	  }

	  if ($state == 2) { # address line 1
	      $address = _trim($_);
	      $state = 3;
	      last SWITCH;
	  }

	  if ($state == 3) { # address line 2 and country code
	      if (/$rex_nsi_country/) {
		  $country = $1;
		  $state = 4;
	      }
	      else {
		  $address .= "\$" . _trim($_);
	      }
	      last SWITCH;
	  }

	  if ($state == 4) { # Administrative
	      $state = 5 if /$rex_admin_contact/;
	      last SWITCH;
	  }

	  if ($state == 5) { # ignore the line after Administrative, ...
	      $state = 6;
	      last SWITCH;
	  }

	  if ($state == 6) { # admin
	      $admin = _trim($_);
	      $state = 7;
	      last SWITCH;
	  }

	  if ($state == 7) { # phone and fax
	      if (/$rex_tierra_fax/) {
		  $fax = $1;
		  $state = 8;
	      }
	      elsif (/$rex_tierra_phone/) {
		  $phone = $1;
	      }
	      last SWITCH;
	  }

	  if ($state == 8) { # email
	      if (/$rex_email/) {
		  $email = $1;
		  $state = 9;
	      }
	      last SWITCH;
	  }

	  if ($state == 9) { # dates
	    SWITCH_1:
	      {
		  $created = str2time($1), last SWITCH_1 if /$rex_tierra_created/;
		  $updated = str2time($1), last SWITCH_1 if /$rex_tierra_updated/;
		  $expires = str2time($1), last SWITCH_1 if /$rex_tierra_expires/;
		  $state = 10,		   last SWITCH_1 if /$rex_nsi_ns_begin/;
	      }
	      last SWITCH;
	  }

	  if ($state == 10) { # name servers
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;
		  
		  $nmservers[$#nmservers+1] = $ref;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: TUCOWS
#
sub _p_tucows {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_registrant/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # first line after Registrant:
	      $registrant = _trim($_);
	      $state = 2;
	      last SWITCH;
	  }

	  if ($state == 2) { # address line 1
	      $address = _trim($_);
	      $state = 3;
	      last SWITCH;
	  }

	  if ($state == 3) { # address line 2
	      $address .= "\$" . _trim($_);
	      $state = 35;
	      last SWITCH;
	  }

	  if ($state == 35) { # address line 3 or country code
	      if (/$rex_nsi_country/) {
		  $country = $1;
		  $state = 5;
	      }
	      else {
		  $address .= "\$" . _trim($_);
		  $state = 4;
	      }
	      last SWITCH;
	  }

	  if ($state == 4) { # country code
	      $country = _trim($_);
	      $state = 5;
	      last SWITCH;
	  }

	  if ($state == 5) { # Administrative contact
	      $state = 6 if /$rex_admin_contact/;
	      last SWITCH;
	  }

	  if ($state == 6) { # admin name and email
	      if (/$rex_han_admin/) {
		  $admin = _trim($1);
		  $email = $2;
		  $state = 7;
	      }
	      last SWITCH;
	  }

	  if ($state == 7) { # telephone
	      if (/$rex_tierra_phone/) {
		  $phone = $1;
		  $state = 8;
	      }
	      last SWITCH;
	  }

	  if ($state == 8) { # dates
	    SWITCH_1:
	      {
		  $created = str2time($1), last SWITCH_1 if /$rex_nsi_created/;
		  $updated = str2time($1), last SWITCH_1 if /$rex_nsi_updated/;
		  $expires = str2time($1), last SWITCH_1 if /$rex_nsi_expires/;
		  $state = 9,              last SWITCH_1 if /$rex_nsi_ns_begin/;
	      }
	      last SWITCH;
	  }

	  if ($state == 9) { # name servers
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;

		  $nmservers[$#nmservers+1] = $ref;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# PARSER: YESNIC
#
sub _p_yesnic {
    my $lines = shift;
    my ($domain, $domain_handle);
    my ($registrant, $address, $country);
    my ($admin, $admin_handle, $email, $phone, $fax);
    my ($created, $updated, $expires);
    my ($registrar, $whois_server, $referral, @nmservers);
    my $state = 0;

    foreach (@$lines) {
      SWITCH: {

	  if ($state == 0) { # before Registrant:
	    SWITCH_1: 
	      {
		  $domain	= $1,	last SWITCH_1 if /$rex_domain/;
		  $registrar	= $1,	last SWITCH_1 if /$rex_registrar/;
		  $whois_server	= $1,	last SWITCH_1 if /$rex_server/;
		  $referral	= $1,	last SWITCH_1 if /$rex_referral/;
		  $state	= 1,	last SWITCH_1 if /$rex_yesnic_reg/;
	      }
	      last SWITCH;
	  }

	  if ($state == 1) { # lines after ::Registrant::
	    SWITCH_1:
	      {
		  $registrant = $1,      last SWITCH_1 if /$rex_yesnic_org/;
		  $address    = $1,      last SWITCH_1 if /$rex_yesnic_address/;
		  $address   .= "\$".$1, last SWITCH_1 if /$rex_yesnic_zip/;
		  $country    = $1,      last SWITCH_1 if /$rex_yesnic_country/;
		  $state      = 2,       last SWITCH_1 if /$rex_yesnic_adm/;
	      }
	      last SWITCH;
	  }

	  if ($state == 2) { # lines after ::Administrative Contact::
	    SWITCH_1:
	      {
		  $admin = $1, last SWITCH_1 if /$rex_yesnic_admin/;
		  $email = $1, last SWITCH_1 if /$rex_yesnic_email/;
		  $phone = $1, last SWITCH_1 if /$rex_yesnic_phone/;
		  $fax   = $1, last SWITCH_1 if /$rex_yesnic_fax/;
		  $state = 3,  last SWITCH_1 if /$rex_yesnic_ns_begin/;
	      }
	      last SWITCH;
	  }

	  if ($state == 3) { # lines afetr ::Name Servers::
	      if (/$rex_nsi_ns/) {
		  my $ref = { host => $1 };
		  $ref->{ip} = $2 if $2;
		  
		  $nmservers[$#nmservers+1] = $ref;
	      }
	      elsif (/$rex_yesnic_date/) {
		  $state = 4;
	      }
	      last SWITCH;
	  }

	  if ($state == 4) { # lines after ::Dates
	    SWITCH_1:
	      {
		  $created = str2time($1), last SWITCH_1 if /$rex_yesnic_created/;
		  $updated = str2time($1), last SWITCH_1 if /$rex_yesnic_updated/;
		  $expires = str2time($1), last SWITCH_1 if /$rex_yesnic_expires/;
	      }
	      last SWITCH;
	  }
      }
    }

    my $rv = {
	domain		=> $domain,
	domain_handle	=> $domain_handle,
	registrant	=> $registrant,
	address		=> $address,
	country		=> $country,
	admin		=> $admin,
	admin_handle	=> $admin_handle,
	email		=> $email,
	phone		=> $phone,
	fax		=> $fax,
	created		=> $created,
	updated		=> $updated,
	expires		=> $expires,
	registrar	=> $registrar,
	whois_server	=> $whois_server,
	referral	=> $referral,
	name_servers	=> \@nmservers,
    };

    return $rv;
}

##############################################################################
# Done
#
1;
__END__

##############################################################################
# Documentation.
#

=head1 NAME

Whois - Get and parse Whois database

=head1 SYNOPSIS

  use Whois;

  $wh = new Whois($domain [, { verbose      => 0, 
                               debug        => 0,
                               max_conn_try => 3 }]);

To connect to the Whois server:
  my $wh = new Whois($domain) or die "Can't connect $domain\n";

  die "No match for $domain\n" unless $wh->ok;

For verbose output:
  my $wh = new Whois($domain, { verbose => 1 });

A debugging level can be set as:
  0 : no debug output
  1 : program execution flow
  2 : network connection
  3 : whois parser
  4 : buffer contents

  my $wh = new Whois($domain, { debug => 0 });

Note that greater levels include lower levels.

All fields except domain may be undef.

  print "Domain Name: ",	$wh->domain;
  print "Domain Handle: ",	$wh->domain_handle;
  print "Registrant: ",		$wh->registrant;
  print "Address:\n";
  printf "  %s\n", $_ foreach split(/\$/, $wh->address);
  print "Country: ",		$wh->country;
  print "Administrator: ",	$wh->admin;
  print "Administrator Handle: ",$wh->admin_handle;
  print "Email: ",		$wh->email;
  print "Phone Number: ",	$wh->phone;
  print "Fax Number: ",		$wh->fax;
  print "Created: ",		$wh->created;
  print "Updated: ",		$wh->updated;
  print "Expires: ",		$wh->expires;
  print "Registrar: ",		$wh->registrar;
  print "Whois Server: ",	$wh->whois_server;
  print "Referral: ",		$wh->referral;
  print "Name Servers:\n";
  printf "  %s [%s]\n", $_->{host}, $_->{ip} foreach @{$wh->name_servers};

Note that the address field uses `$' as a new line seperator.

=head1 DESCRIPTION

Blah blah blah.

=head1 AUTHOR

Leo Kim. <leo@florida.sarang.net>

=head1 BUGS

Still this module is considered an alpha quality, which means there are lots of bugs.

=head1 SEE ALSO

blah

=cut

#
# Local Variables:
# mode: perl
# change-log-default-name: Changes
# End:
#
# -*- End-Of-Module -*-
